;; Generated package description from detached.el  -*- no-byte-compile: t -*-
(define-package "detached" "0.9.2.0.20221118.122609" "A package to launch, and manage, detached processes" '((emacs "27.1")) :commit "8c7a2e1129b254946d86a7b7ae1c8378afab7b69" :authors '(("Niklas Eklund" . "niklas.eklund@posteo.net")) :maintainer '("detached.el Development" . "~niklaseklund/detached.el@lists.sr.ht") :keywords '("convenience" "processes") :url "https://sr.ht/~niklaseklund/detached.el/")
